const States = new Map();
const Refs = new Map();
const Functions = new Map();
const EventHandlers = new Map();
const Parts = new Set();
const Conditions = new Set();
const Lists = new Set();

function State(initialValue) {
    let stateValue = initialValue;
    let stateId = `SkyState-${Date.now() * Math.floor(Math.random() * 10000000)}`;
    States.set(stateId, stateValue);

    function updateState(value) {
        stateValue = typeof value === "function" ? value(stateValue) : value;
        if (Functions.has(stateId)) Functions.get(stateId)(stateValue);
        States.set(stateId, stateValue);

        /* Conditional Rendering */
        [...Conditions.values()].forEach((Part) => {
            if (!(Part[0] instanceof HTMLElement)) return;
            let Replaces = Part[1]();
            if (Replaces === Part[0].textContent) return;
            Part[0].innerHTML = Replaces;
        });

        /* Reactive Parts */
        [...Parts.values()].forEach((Part) => {
            if (!Part[1].includes(stateId)) return;
            let Replaces = Part[1].replaceAll(stateId, stateValue);
            if (Part[0].nodeType === 2) {
                Part[1].match(/SkyState\-\d+/g).forEach((match) => {
                    Replaces = Replaces.replaceAll(match, States.get(match));
                });
                Part[0].value = Replaces;
                return;
            };
            Part[1].match(/SkyState\-\d+/g).forEach((match) => {
                Replaces = Replaces.replaceAll(match, States.get(match));
            });

            Part[0].textContent = Replaces;
        });

        /* List Rendering */
        if (stateValue instanceof Array) [...Lists.values()].forEach((List) => {
            if (!(List instanceof Array) || List[1] !== stateId) return;
            [...List[0].children].forEach(child => child.remove())
            let doc = "";
            stateValue.forEach((ele) => doc += List[2](ele));
            List[0].innerHTML = doc;
        });
    };

    function onStateUpdate(func) {
        if (typeof func !== "function") return;
        Functions.set(stateId, func);
    };
    return [(id) => id?.toLowerCase().trim() === "id" ? stateId : stateValue, updateState, onStateUpdate];
};

function Ref() {
    let refId = `SkyRef-${Date.now() * Math.floor(Math.random() * 10000000)}`;
    Refs.set(refId, [null]);

    function onReference(func) {
        if (typeof func !== "function") return;
        Refs.set(refId, [null, func]);
    }
    return [() => Refs.get(refId)[0] || refId, onReference]
}

function Document(htmlString, parentElement) {
    if (htmlString instanceof Array) htmlString = htmlString[0];
    if (typeof htmlString !== "string") throw new Error("Invalid HTML string");
    const Parser = new DOMParser();
    const doc = Parser.parseFromString(htmlString.replaceAll(/\<\#(\s+[^>]*?)?>/g, "<SkyWraper$1>").replaceAll("</#>", "</SkyWraper>"), "text/html");
    let root = [...doc.documentElement.querySelector("body").children].length === 1 ? [doc.documentElement.querySelector("body").children[0], ...doc.documentElement.querySelector("body").children[0].children] : null;
    if (root === null) throw Error("Document elements must be grouped by a wrapper element or use \"<#></#>\"");
    let children = allChildren({ children: root });
    let newRoot = doc.documentElement.querySelector("body").children[0];
    if (parentElement) {
        if (typeof parentElement === "string" && document.querySelector(parentElement) instanceof HTMLElement) {
            Render(children);
            document.querySelector(parentElement).appendChild(newRoot);
        };
        if (parentElement instanceof HTMLElement) {
            Render(children);
            parentElement.appendChild(newRoot);
        };
        if (parentElement instanceof HTMLCollection) {
            [...parentElement].forEach((element) => {
                let root = newRoot.cloneNode(true);
                let children = allChildren({ children: [root, ...root.children] });
                Render(children);
                element.appendChild(root);
            });
        };
        if (parentElement instanceof NodeList) {
            [...parentElement].forEach((element) => {
                let root = newRoot.cloneNode(true);
                let children = allChildren({ children: [root, ...root.children] });
                Render(children);
                element.appendChild(root);
            });
        };
    } else Render(elements);
    if (newRoot === null) throw Error("Document elements must be grouped by a wrapper element or use \"<#></#>\"")
    return newRoot;
};

function allChildren(element) {
    let elements = [...element.children];
    for (let i = 0; i < element.children.length; i++) {
        if (element.children[i].children) elements = [...elements, ...allChildren({ children: element.children[i].children })]
    };
    return elements;
};

function Render(elements) {
    for (let element of elements) {
        for (let node of element.childNodes) {
            if (node.nodeType === 3 && node.textContent.match(/Sky(List)?State\-\d+/g)) {
                Parts.add([node, node.textContent.replaceAll("SkyListState-", "SkyState-")]);
                if (node.textContent.match(/Sky(List)?State\-\d+/g)) node.textContent.match(/Sky(List)?State\-\d+/g).forEach((match) => {
                    node.textContent = node.textContent.replaceAll(match, States.get(match.replaceAll("SkyListState", "SkyState")));
                });
            };
        };
    };
    for (let i = 0; i < elements.length; i++) {
        [...elements[i].attributes].forEach(attr => {
            if (attr.name === "ref") {
                if (!Refs.has(elements[i].getAttribute("ref"))) return;
                let func = Refs.get(elements[i].getAttribute("ref"))[1];
                Refs.set(elements[i].getAttribute("ref"), [elements[i]]);
                if (func) func(elements[i]);
                elements[i].removeAttribute("ref");
            } else if (attr.name.startsWith("on-") && attr.name.split("-")[1] !== "") {
                if (EventHandlers.has(attr.value)) {
                    elements[i].addEventListener(attr.name.split("-")[1].toLowerCase(), EventHandlers.get(attr.value));
                } else elements[i].addEventListener(attr.name.split("-")[1].toLowerCase(), new Function(`return ${attr.value}`)());
                elements[i].removeAttribute(attr.name);
            } else {
                if (attr.value.match(/SkyState\-\d+/g)) {
                    Parts.add([attr, attr.value]);
                    if (attr.value.match(/SkyState\-\d+/g)) attr.value.match(/SkyState\-\d+/g).forEach((match) => {
                        attr.value = attr.value.replaceAll(match, States.get(match));
                    });
                };
            }
        });
    };
};

function ForEach(list, callback) {
    let listId = null;
    if (typeof list === "string" && (list.match(/SkyState\-\d+/g)[0] && list === list.match(/SkyState\-\d+/g)[0])) {
        listId = list.match(/SkyState\-\d+/g)[0];
        list = States.get(listId);
    };
    if (!(list instanceof Array)) {
        // Maybe adding Object
        throw Error("List is not an array");
    };
    if (typeof callback !== "function") {
        throw Error("Callback is not a function");
    };
    let [ref, onRef] = Ref();
    let refId = ref();
    let doc = "";
    onRef((element) => {
        let parent = element.parentElement;
        parent.innerHTML = doc;
        if (listId) Lists.add([parent, listId, callback]);
        else {
            Render(allChildren(parent));
        };
        Refs.delete(refId);
    });
    list.forEach((ele) => doc += callback(ele));
    return `<holder ref="${ref()}"></holder>`;
};

function While(condition, callback, delay) {
    if (typeof condition !== "function") {
        throw Error("Condition is not a function");
    };
    if (typeof callback !== "function") {
        throw Error("Callback is not a function");
    };
    let [ref, onRef] = Ref();
    let refId = ref();
    let doc = "";
    let parent;
    onRef((element) => {
        parent = element.parentElement;
        if (typeof delay !== "number" || delay < 0) parent.innerHTML = doc;
        Refs.delete(refId);
    });
    if (typeof delay === "number" && delay > 0) {
        let Loop = setInterval(() => {
            if (!condition()) return clearInterval(Loop);
            doc += callback();
            parent.innerHTML = doc;
        }, delay);
    } else while (condition()) {
        doc += callback();
    };
    return `<holder ref="${ref()}"></holder>`;
};

function Condition(callback) {
    if (typeof callback !== "function") return;
    let [ref, onRef] = Ref();
    let refId = ref();
    let condition = callback();
    onRef((element) => {
        let parent = element.parentElement;
        parent.innerHTML = condition ?? "";
        Conditions.add([parent, callback]);
        Refs.delete(refId);
    });
    return `<holder ref="${ref()}"></holder>`;
};

function EventHandler(callback) {
    let eventHandlerId = `SkyEventHandler-${Date.now() * Math.floor(Math.random() * 10000000)}`;
    EventHandlers.set(eventHandlerId, callback);
    return eventHandlerId;
};